"""
Data attributes for events within the architecture subdomain `learning`.

These attributes follow the form of attr objects specified in OEP-49 data
pattern.
"""
from datetime import datetime
from typing import List, Optional

import attr
from opaque_keys.edx.keys import CourseKey, UsageKey


@attr.s(frozen=True)
class UserNonPersonalData:
    """
    Attributes defined for Open edX user object based on non-PII data.

    Arguments:
        id (int): unique identifier for the Django User object.
        is_active (bool): indicates whether the user is active.
    """

    id = attr.ib(type=int)
    is_active = attr.ib(type=bool)


@attr.s(frozen=True)
class UserPersonalData:
    """
    Attributes defined for Open edX user object based on PII data.

    Arguments:
        username (str): username associated with the Open edX user.
        email (str): email associated with the Open edX user.
        name (str): name associated with the Open edX user's profile.
    """

    username = attr.ib(type=str)
    email = attr.ib(type=str)
    name = attr.ib(type=str, factory=str)


@attr.s(frozen=True)
class UserData(UserNonPersonalData):
    """
    Attributes defined for Open edX user object.

    This class extends UserNonPersonalData to include PII data completing the
    user object.

    Arguments:
        pii (UserPersonalData): user's Personal Identifiable Information.
    """

    pii = attr.ib(type=UserPersonalData)


@attr.s(frozen=True)
class CourseData:
    """
    Attributes defined for Open edX Course Overview object.

    Arguments:
        course_key (str): identifier of the Course object.
        display_name (str): display name associated with the course.
        start (datetime): start date for the course.
        end (datetime): end date for the course.
    """

    course_key = attr.ib(type=CourseKey)
    display_name = attr.ib(type=str, factory=str)
    start = attr.ib(type=datetime, default=None)
    end = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class CourseEnrollmentData:
    """
    Attributes defined for Open edX Course Enrollment object.

    Arguments:
        user (UserData): user associated with the Course Enrollment.
        course (CourseData): course where the user is enrolled in.
        mode (str): course mode associated with the course enrollment.
        is_active (bool): whether the enrollment is active.
        creation_date (datetime): creation date of the enrollment.
        created_by (UserData): if available, who created the enrollment.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    mode = attr.ib(type=str)
    is_active = attr.ib(type=bool)
    creation_date = attr.ib(type=datetime)
    created_by = attr.ib(type=UserData, default=None)


@attr.s(frozen=True)
class CertificateData:
    """
    Attributes defined for Open edX Certificate data object.

    Arguments:
        user (UserData): user associated with the Certificate.
        course (CourseData): course where the user obtained the certificate.
        mode (str): course mode associated with the course enrollment.
        grade (str): user's grade in this course run.
        download_url (str): URL where the PDF version of the certificate.
        name (str): user's name.
        current_status (str): current certificate status.
        previous_status (str): if available, pre-event certificate status.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    mode = attr.ib(type=str)
    grade = attr.ib(type=str)
    download_url = attr.ib(type=str)
    name = attr.ib(type=str)
    current_status = attr.ib(type=str)
    previous_status = attr.ib(type=str, factory=str)


@attr.s(frozen=True)
class CohortData:
    """
    Attributes defined for Open edX Cohort Membership object.

    Arguments:
        user (UserData): user assigned to the group.
        course (CourseData): course associated with the course group.
        name (str): name of the cohort group.
    """

    user = attr.ib(type=UserData)
    course = attr.ib(type=CourseData)
    name = attr.ib(type=str)


@attr.s(frozen=True)
class DiscussionTopicContext:
    """
    Attributes defined for Open edX Discussion Topic Context object.

    Context for linking the external id for a discussion topic to its associated usage key.

    Arguments:
        title (str): title of the discussion. This field is cached to improve the performance, since otherwise we'd
        need to look it up in the course structure each time.
        usage_key (str): unit location.
        group_id (Optional[int]): can be used for providers that don't internally support
        cohorting but we can emulate that with different contexts for different groups.
        external_id (str): store the commentable id that is used by cs_comments_service.
        ordering (int): represent the position of the discussion topic.
        context (dict): additional structured information about the context in
          which this topic is used, such as the section, subsection etc.
    """

    title = attr.ib(type=str)
    usage_key = attr.ib(type=UsageKey, default=None)
    group_id = attr.ib(type=Optional[int], default=None)
    external_id = attr.ib(type=str, default=None)
    ordering = attr.ib(type=int, default=None)
    context = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class CourseDiscussionConfigurationData:
    """
    Attributes defined for Open edX Course Discussion Configuration Data object.

    Course configuration information for discussions. Contains all the metadata
    needed to configure discussions for a course.

    Arguments:
        course_key (str): identifier of the course to which the discussion belongs.
        provider_type (str): provider type from discussion settings.
        enable_in_context (bool): indicates whether in-context discussion is enabled for the course
        enable_graded_units (bool): If enabled, discussion topics will be created for graded units as well.
        unit_level_visibility (bool): visibility for unit level.
        plugin_configuration (dict): The plugin configuration data for this context/provider.
        contexts (List[DiscussionTopicContext]): contains all the contexts for which discussion
        is to be enabled.
    """

    course_key = attr.ib(type=CourseKey)
    provider_type = attr.ib(type=str)
    enable_in_context = attr.ib(type=bool, default=True)
    enable_graded_units = attr.ib(type=bool, default=False)
    unit_level_visibility = attr.ib(type=bool, default=False)
    plugin_configuration = attr.ib(type=dict, default={})
    contexts = attr.ib(type=List[DiscussionTopicContext], factory=list)


@attr.s(frozen=True)
class PersistentCourseGradeData:
    """
    Attributes defined for Open edX PersistentCourseGrade data object.

    Arguments:
        user_id (int): identifier of the grade to which the grade belongs.
        course (CourseData): Identifier of the course to which the grade belongs.
        course_edited_timestamp (datetime): date the course was edited.
        course_version (str): version of the course.
        grading_policy_hash (str): grading policy hash of the course.
        percent_grade (float): percentage of the grade.
        letter_grade (str): grade in letter
        passed_timestamp (datetime): date the course was passed.
    """

    user_id = attr.ib(type=int)
    course = attr.ib(type=CourseData)
    course_edited_timestamp = attr.ib(type=datetime)
    course_version = attr.ib(type=str)
    grading_policy_hash = attr.ib(type=str)
    percent_grade = attr.ib(type=float)
    letter_grade = attr.ib(type=str)
    passed_timestamp = attr.ib(type=datetime)


@attr.s(frozen=True)
class XBlockSkillVerificationData:
    """
    Data needed to update verification count  of tags/skills for an XBlock.

    User feedback on whether tags/skills related to an XBlock are valid.

    Arguments:
        usage_key (UsageKey): identifier of the XBlock object.
        verified_skills (List[int]): list of verified skill ids.
        ignored_skills (List[int]): list of ignored skill ids.
    """

    usage_key = attr.ib(type=UsageKey)
    verified_skills = attr.ib(type=List[int], factory=list)
    ignored_skills = attr.ib(type=List[int], factory=list)


@attr.s(frozen=True)
class UserNotificationData:
    """
    Attributes defined for Open edX User Notification data object.

    Arguments:
        user_ids (List(int)): identifier of the users to which the notification belongs.
        notification_type (str): type of the notification.
        content_url (str): url of the content.
        app_name (str): name of the app.
        course_key (str): identifier of the Course object.
        context (dict): additional structured information about the context of the notification.
    """

    user_ids = attr.ib(type=List[int])
    notification_type = attr.ib(type=str)
    content_url = attr.ib(type=str)
    app_name = attr.ib(type=str)
    course_key = attr.ib(type=CourseKey)
    context = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class ProgramData:
    """
    Attributes defined for the Open edX Program data object.

    Arguments:
        uuid (str): The UUID of the program (from Course-Discovery)
        title (str): The title of the program
        program_type (str): The type slug of the program (e.g. professional, microbachelors, micromasters, etc.)
    """

    uuid = attr.ib(type=str)
    title = attr.ib(type=str)
    program_type = attr.ib(type=str)


@attr.s(frozen=True)
class ProgramCertificateData:
    """
    Attributes defined for the Open edX Program Certificate data object.

    Arguments:
        user (UserData): User associated with the Program Certificate
        program (ProgramData): Program data associated with the Program Certificate
        uuid (str): UUID of the UserCredential record in Credentials
        certificate_available_date (datetime): Optional. A DateTime describing when a learner is allowed to view the
                                                credential
        status (str): The status of the credential (e.g. `awarded` or `revoked`)
        url (str): A URL to the learner's credential
    """

    user = attr.ib(type=UserData)
    program = attr.ib(type=ProgramData)
    uuid = attr.ib(type=str)
    status = attr.ib(type=str)
    url = attr.ib(type=str)
    certificate_available_date = attr.ib(type=datetime, default=None)


@attr.s(frozen=True)
class ExamAttemptData:
    """
    Attributes defined for the Open edX Exam downstream effects.

    Note that events that use this data type:
        A. Pretain to "Special Exams", e.g. Timed or Proctored exams, and not non-timed course
        subsections that are labelled as an exam.
        B. Are only ever emitted from the newer exams backend, edx-exams, and never from the
        legacy exams backend, edx-proctoring.

    The event signals that use this data have the prefix `EXAM_`, which is equivalent to "special exam".
    We are using this as a shortened form of the prefix `SPECIAL_EXAM` to avoid confusion, as these are likely
    the only type of exams that will be of concern in the context of events/the event bus.

    For more information, please see the ADR relating to this event data:
    https://github.com/openedx/openedx-events/blob/main/docs/decisions/0013-special-exam-submission-and-review-events.rst

    Arguments:
        student_user (UserData): user object for the student to whom the exam attempt belongs
        course_key (CourseKey): identifier of the course to which the exam attempt belongs
        usage_key (UsageKey): identifier of the content that will get a exam attempt
        exam_type (str): type of exam that was taken (e.g. timed, proctored, etc.)
        requesting_user (UserData): user triggering the event (sometimes a non-learner, e.g. an instructor)
    """

    student_user = attr.ib(type=UserData)
    course_key = attr.ib(type=CourseKey)
    usage_key = attr.ib(type=UsageKey)
    exam_type = attr.ib(type=str)
    requesting_user = attr.ib(type=UserData, default=None)


@attr.s(frozen=True)
class CourseAccessRoleData:
    """
    Attributes defined for the Open edX Course Access Role data object.

    Arguments:
        user (UserData): user associated with the CourseAccessRole.
        course_key (CourseKey): identifer of the related course object.
        org (str): identifier of the organization.
        role (str): the role of the user in the course.
    """

    user = attr.ib(type=UserData)
    org_key = attr.ib(type=str)
    course_key = attr.ib(type=CourseKey)
    role = attr.ib(type=str)


@attr.s(frozen=True)
class ManageStudentsPermissionData:
    """
    Attributes defined for the Open edX to represent users that can manage students within a course/org.

    IMPORTANT:
        edX currently uses roles, and only roles, to decide what kind of access a user has.

        There is an ongoing project to replace this roles-only system with a system that uses roles that are
        made up of permissions, which is being worked on in parallel with another project to emit events
        whenever users are assigned any type of "Course Staff" role.

        It's unclear what the state of this roles/permissions project will be the time the events project
        is completed, so each project's respective teams will stay in touch with each other.

        For now, we're making a best effort to publish this an event that will regard the permission(s)
        we'd expect to "filter" for in the future (For more info, please check out this document:
        https://docs.google.com/spreadsheets/d/1htsV0eWq5-y96DZ5A245ukfZ4_qeH0KjHVaOyfqD8OA/edit#gid=908503896)
        and not for the roles we have now. Likely this/these permission(s) will be something like `manage_students`,
        but we need to evaluate how this will align with some possible future roles such as `limited_staff` or `ccx`.

        As such, the current plan is to do one of the following once the roles/permissions project's
        feature branch is merged to master:
            1. Modify this event to "filter" by the correct permissions once the
            2. As a backup plan, make a new event if this proves too difficult.

        Until either of these plans are executed, this comment under the IMPORTANT header should stay put.

    Arguments:
        user (UserData): User who will have a role/permission assigned/removed.
        permission (str): The permission the user is being assigned.
        course_key (Course ID): identifier of the course where the user will have staff role assigned/removed.
            A blank course_id implies org wide role.
        org (str): identifier of the org where the user will have staff role assigned/removed.
            A blank org is for global group based roles such as course creator (may be deprecated).
    """

    user = attr.ib(type=UserData)
    permission = attr.ib(type=str)
    course_key = attr.ib(type=str, default=None)
    org = attr.ib(type=str, default=None)


@attr.s(frozen=True)
class DiscussionThreadData:
    """
    Attributes defined for the Open edX to represent events in the Forum such as comments, responses, and threads.

    For more details on the data attributes, please see the following documentation:
    https://docs.openedx.org/en/latest/developers/references/internal_data_formats/tracking_logs/student_event_types.html#edx-forum-thread-created

    Arguments:
        anonymous (bool): indicates whether the user is anonymous.
        anonymous_to_peers (bool): indicates whether the user is anonymous to peers.
        body (str): body of the discussion thread.
        category_id (int): identifier of the category.
        category_name (str): name of the category.
        commentable_id (str): identifier of the commentable.
        group_id (int): identifier of the group.
        id (int): identifier of the discussion thread.
        team_id (int): identifier of the team.
        thread_type (str): type of the thread.
        title (str): title of the thread.
        title_truncated (bool): indicates whether the title is truncated.
        truncated (bool): indicates whether the thread is truncated.
        url (str): url of the thread.
        user (UserData): information of the user that authored the thread/comment/response.
        course_id (CourseKey): identifier of the course.
        discussion (dict): discussion data. (optional, specific to comments and responses)
        user_course_roles (List[str]): user course roles.
        user_forums_roles (List[str]): user forums roles.
        options (dict): options for the thread.
    """

    anonymous = attr.ib(type=bool)
    anonymous_to_peers = attr.ib(type=bool)
    body = attr.ib(type=str)
    category_id = attr.ib(type=int)
    category_name = attr.ib(type=str)
    commentable_id = attr.ib(type=str)
    group_id = attr.ib(type=int)
    id = attr.ib(type=int)
    team_id = attr.ib(type=int)
    thread_type = attr.ib(type=str)
    title = attr.ib(type=str)
    title_truncated = attr.ib(type=bool)
    truncated = attr.ib(type=bool)
    url = attr.ib(type=str)
    user = attr.ib(type=UserData)
    course_id = attr.ib(type=CourseKey)
    discussion = attr.ib(type=dict, factory=dict)
    user_course_roles = attr.ib(type=List[str], factory=list)
    user_forums_roles = attr.ib(type=List[str], factory=list)
    options = attr.ib(type=dict, factory=dict)


@attr.s(frozen=True)
class CourseNotificationData:
    """
    Attributes defined for Open edX Course Notification data object.

    Arguments:
        course_key (str): identifier of the Course object.
        app_name (str): name of the app requesting the course notification.
        notification_type (str): type of the notification.
        content_url (str): url of the content the notification will redirect to.
        content_context (dict): additional information related to the content of the notification.
            Notification content templates are defined in edx-platform here:
                https://github.com/openedx/edx-platform/blob/master/openedx/core/djangoapps/notifications/base_notification.py#L10

        Example of content_context for a discussion notification (new_comment_on_response):

            {
                ...,
                "content_context": {
                    "post_title": "Post Title",
                    "replier_name": "test_user",
            }

        audience_filters (dict): additional information related to the audience of the notification.
            We can have different filters on course level, such as roles, enrollments, cohorts etc.

        Example of audience_filters for a discussion notification (new_discussion_post):

            {
                ...,
                "audience_filters": {
                    "enrollment": ["verified", "audit"],
                    "role": ["discussion admin", "discussion moderator"],
            }
    """

    course_key = attr.ib(type=CourseKey)
    app_name = attr.ib(type=str)
    notification_type = attr.ib(type=str)
    content_url = attr.ib(type=str)
    content_context = attr.ib(type=dict, factory=dict)
    audience_filters = attr.ib(type=dict, factory=dict)



#!/usr/bin/env python
"""
Package metadata for openedx_events.
"""
import os
import re
import sys

from setuptools import setup


def get_version(*file_paths):
    """
    Extract the version string from the file.

    Input:
     - file_paths: relative path fragments to file with
                   version string
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename) as file:
        version_file = file.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.

    Returns:
        list: Requirements file relative path strings
    """
    requirements = set()
    for path in requirements_paths:
        with open(path) as file:
            requirements.update(
                line.split('#')[0].strip() for line in file.readlines()
                if is_requirement(line.strip())
            )
    return list(requirements)


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.

    Returns:
        bool: True if the line is not blank, a comment, a URL, or
              an included file
    """
    return line and not line.startswith(('-r', '#', '-e', 'git+', '-c'))


VERSION = get_version('openedx_events', '__init__.py')

if sys.argv[-1] == 'tag':
    print("Tagging the version on github:")
    os.system("git tag -a %s -m 'version %s'" % (VERSION, VERSION))
    os.system("git push --tags")
    sys.exit()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()
with open(os.path.join(os.path.dirname(__file__), 'CHANGELOG.rst')) as changelog:
    CHANGELOG = changelog.read()

setup(
    name='openedx-events',
    version=VERSION,
    description="""Open edX events from the Hooks Extensions Framework""",
    long_description=README + '\n\n' + CHANGELOG,
    long_description_content_type='text/x-rst',
    author='edX',
    author_email='oscm@edx.org',
    url='https://github.com/openedx/openedx-events',
    packages=[
        'openedx_events',
    ],
    include_package_data=True,
    install_requires=load_requirements('requirements/base.in'),
    python_requires=">=3.8",
    license="Apache 2.0",
    zip_safe=False,
    keywords='Python edx',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
)



"""
Serialize events to Avro records.
"""
import io
import json

import attr
import fastavro

from .custom_serializers import DEFAULT_CUSTOM_SERIALIZERS
from .schema import schema_from_signal

DEFAULT_SERIALIZERS = {serializer.cls: serializer.serialize for serializer in DEFAULT_CUSTOM_SERIALIZERS}


def _get_non_attrs_serializer(serializers=None):
    """
    Create a method to pass as the value_serializer argument to attr.as_dict to serialize using custom serializers.

    Arguments:
        serializers: A map of Python type to serialization method

    Returns:
        A method for serializing non_attrs values
    """
    param_serializers = serializers or {}
    all_serializers = {**DEFAULT_SERIALIZERS, **param_serializers}

    def _serialize_non_attrs_values(inst, field, value):  # pylint: disable=unused-argument
        if value is None:
            # All default=None fields are implicit union types of NoneType
            # and something else. (See ADR 7.) Note that if there isn't a
            # default at all, field.default will be attrs.NOTHING, not None.
            if field.default is None:
                return None
            else:
                # Bail out early with an informative message rather
                # than ending up with an inscrutable error from inside
                # a custom serializer.
                #
                # If we ever make a custom serializer that can handle
                # None as an input, we can remove this check.
                # pylint: disable-next=broad-exception-raised
                raise Exception("None cannot be handled by custom serializers (and default=None was not set)")

        for extended_class, serializer in all_serializers.items():
            if field:
                # Make sure that field.type is a class first.
                if isinstance(field.type, type) and issubclass(field.type, extended_class):
                    return serializer(value)
            if issubclass(type(value), extended_class):
                return serializer(value)
        return value
    return _serialize_non_attrs_values


def _event_data_to_avro_record_dict(event_data, serializers=None):
    """
    Create an Avro record dictionary from an event data dict.

    Arguments:
        event_data: A dictionary representing an event sent by an instance of OpenEdxPublicSignal
        serializers: A map of Python type to serialization method

    Returns:
        An Avro record dictionary representation of the event data
    """

    def value_to_dict(value):
        # Case 1: Value is an instance of an attrs-decorated class
        if hasattr(value, "__attrs_attrs__"):
            return attr.asdict(value, value_serializer=_get_non_attrs_serializer(serializers))
        return _get_non_attrs_serializer(serializers)(None, None, value)

    return json.loads(
        json.dumps(
            event_data, sort_keys=True, default=value_to_dict
        )
    )


def serialize_event_data_to_bytes(event_data, signal):
    """
    Serialize event data to bytes.

    Arguments:
        event_data: Event data to be sent via an OpenEdxPublicSignal's send_event method
        signal: An instance of OpenEdxPublicSignal
    Returns:
        bytes: Byte representation of the event_data, to be sent over the wire
    """
    serializer = AvroSignalSerializer(signal)
    schema_dict = serializer.schema
    out = io.BytesIO()
    data_dict = serializer.to_dict(event_data)
    fastavro.schemaless_writer(out, schema_dict, data_dict)
    out.seek(0)
    return out.read()


class AvroSignalSerializer:
    """
    Class to serialize event data dictionaries into Avro record dictionaries that can be sent by an event bus.

    The ``schema_string`` and ``to_dict`` methods are the API for this deserializer. This API is derived from the
    confluent_kafka.AvroSerializer class, which is part of the Kafka event bus ecosystem. The AvroSerializer takes a
    schema (as string) and a to_dict method as initialization parameters. These methods could also potentially be used
    by other event bus implementations.

    To serialize events that include data types that are not yet supported, see README.
    """

    def __init__(self, signal):
        """
        Initialize serializer, creating an Avro schema from signal.

        Arguments:
            signal: An instance of OpenEdxPublicSignal
        """
        self.signal = signal
        self.serializers = {ext.cls: ext.serialize for ext in self.custom_type_serializers()}
        self.custom_types = {ext.cls: ext.field_type for ext in self.custom_type_serializers()}
        self.schema = schema_from_signal(self.signal, custom_type_to_avro_type=self.custom_types)

    def schema_string(self):
        """Get Avro schema as JSON string."""
        return json.dumps(self.schema, sort_keys=True)

    def to_dict(self, event_data):
        """Convert event data to an Avro record dictionary."""
        return _event_data_to_avro_record_dict(event_data, serializers=self.serializers)

    def custom_type_serializers(self):
        """
        Override this method to add custom serializers for unhandled classes.

        Returns:
            A list of subclasses of BaseCustomTypeAvroSerializer
        """
        return []



"""
Tests for consume_events command.
"""
from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase

from openedx_events.management.commands.consume_events import Command


class TestCommand(TestCase):
    """
    Tests for the consume_events management command
    """

    @patch('openedx_events.tooling.OpenEdxPublicSignal.get_signal_by_type', return_value="test-signal")
    @patch('openedx_events.management.commands.consume_events.make_single_consumer', autospec=True)
    def test_consumer_call_with_only_required_args(self, mock_make_consumer, _):
        """
        This methods checks the required args are correctly passed to consumer.

        Expected behavior:
            The required args are passed to consumer.
        """
        call_command(Command(), topic=['test'], group_id=['test'])
        mock_make_consumer.assert_called_once_with(topic='test', group_id='test')

    @patch('openedx_events.management.commands.consume_events.make_single_consumer', autospec=True)
    def test_consumer_call_with_extra_args(self, mock_make_consumer):
        """
        This methods checks the extra args are correctly parsed and passed to consumer.

        Expected behavior:
            The extra args are parsed and passed to consumer.
        """
        call_command(
            Command(),
            topic=['test'],
            group_id=['test'],
            extra='{"check_backlog":true}'
        )
        mock_make_consumer.assert_called_once_with(
            topic='test',
            group_id='test',
            check_backlog=True
        )

    @patch('openedx_events.management.commands.consume_events.logger', autospec=True)
    @patch('openedx_events.management.commands.consume_events.make_single_consumer', autospec=True)
    def test_consumer_call_with_incorrect_json_string(self, mock_make_consumer, mock_logger):
        """
        This methods checks that consumer is not called with incorrect args if extra args json is incorrectly formed.

        Expected behavior:
            The command logs and skips execution if extra args json is incorrectly formed.
        """
        call_command(
            Command(),
            topic=['test'],
            group_id=['test'],
            extra="{'check_backlog':true}"
        )
        mock_logger.exception.assert_called_once_with("Error consuming events")
        mock_make_consumer.assert_not_called()



"""Tests for generate_avro_schemas."""
import os
from importlib import import_module
from unittest.mock import call, mock_open, patch

from django.core.management import call_command
from django.test import TestCase

from openedx_events.event_bus.avro.tests.test_utilities import create_simple_signal
from openedx_events.management.commands.generate_avro_schemas import Command
from openedx_events.tests.utils import FreezeSignalCacheMixin
from openedx_events.tooling import KNOWN_UNSERIALIZABLE_SIGNALS, OpenEdxPublicSignal, load_all_signals


class TestGenerateAvroCommand(FreezeSignalCacheMixin, TestCase):
    """
    Tests for generate_avro_schemas management command.
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.root_path = import_module('openedx_events').__path__[0]
        cls.folder_path = f"{cls.root_path}/event_bus/avro/tests/schemas"

    # assume the schema folder already exists but has no files so we're not overwriting anything
    @patch('os.path.exists', lambda path: path == TestGenerateAvroCommand.folder_path)
    def test_generate_multiple_specified_schemas(self):
        signal_a = create_simple_signal({'key': str})
        signal_b = create_simple_signal({'key': str}, event_type='simple.signal.B')
        with patch("builtins.open", mock_open()) as mock_file:
            call_command(Command(), signal_a.event_type, signal_b.event_type)
            # make sure we created files with the correct name
            mock_file.assert_has_calls([
                    call(f"{TestGenerateAvroCommand.folder_path}/simple+signal_schema.avsc", "w"),
                    call(f"{TestGenerateAvroCommand.folder_path}/simple+signal+B_schema.avsc", "w"),
                ],
                # need any_order=True because opening a file involves a bunch of other secret calls
                any_order=True)
            handle = mock_file()
            # make sure we wrote the schemas. Exact JSON determined from testing
            assert handle.write.mock_calls == [
                call('{\n  "name": "CloudEvent",\n  "type": "record",\n'
                     '  "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",'
                     '\n  "fields": [\n    {\n      "name": "key",\n      "type": "string"\n    }\n  ],\n  "namespace":'
                     ' "simple.signal"\n}'),
                call('{\n  "name": "CloudEvent",\n  "type": "record",\n'
                     '  "doc": "Avro Event Format for CloudEvents created with openedx_events/schema",'
                     '\n  "fields": [\n    {\n      "name": "key",\n      "type": "string"\n    }\n  ],\n  "namespace":'
                     ' "simple.signal.B"\n}')
            ]

    # pretend that all files in the schema folder have already been created
    @patch('os.path.exists', lambda path: TestGenerateAvroCommand.folder_path in path)
    def test_command_warns_if_schema_exists(self):
        signal_a = create_simple_signal({'key': str})
        with patch("builtins.open", mock_open()) as mock_file:
            # test we skip if the user doesn't want to continue
            with patch("builtins.input", return_value='n'):
                call_command(Command(), signal_a.event_type)
                mock_file.assert_not_called()
            # check we continue if the user wants to continue
            with patch("builtins.input", return_value='Y'):
                call_command(Command(), signal_a.event_type)
                mock_file.assert_has_calls([
                    call(f"{TestGenerateAvroCommand.folder_path}/simple+signal_schema.avsc", "w"),
                ])

    @patch('os.path.exists', lambda path: path == TestGenerateAvroCommand.folder_path)
    def test_generate_all(self):
        load_all_signals()
        expected_files = [f"{TestGenerateAvroCommand.folder_path}/{signal.event_type.replace('.','+')}_schema.avsc"
                          for signal in OpenEdxPublicSignal.all_events() if signal.event_type
                          not in KNOWN_UNSERIALIZABLE_SIGNALS]
        with patch("builtins.open", mock_open()) as mock_file:
            call_command(Command(), all=True)
            mock_file.assert_has_calls([call(file, 'w') for file in expected_files], any_order=True)

    # mostly added to appease the coverage gods
    @patch('os.makedirs')
    def test_create_schema_dir_if_not_exists(self, mock_makedirs):
        signal = create_simple_signal({"key": str})
        old_exists = os.path.exists
        with patch('os.path.exists',
                   lambda path: False if path == TestGenerateAvroCommand.folder_path else old_exists(path)):
            with patch("builtins.open", mock_open()):
                call_command(Command(), signal.event_type)
        mock_makedirs.assert_called_once_with(TestGenerateAvroCommand.folder_path)
